/*
 * hdreset
 *
 * Resets hard disk checksum in EEPROM Atmel 24C08 of Hifidelio devices
 * via I2c-Bus. The checksum is located at register 0xf9 of chip address
 * 0x53.
 * 
 * v1.0 - First release.
 * v1.1 - Added command line option -n --nocheck and usage text.
 */

// Standard includes

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include <sys/ioctl.h>

// Kernel includes

#include <linux/hdreg.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>

// Global defines

#define I2C_REGISTER 0xf9
#define I2C_ADDRESS  0x53
#define DEV_HD  "/dev/hda"
#define DEV_I2C "/dev/i2c-0"

// Application name

char *app_name = NULL;


/* 
 * Error handling. Prints error message and exits program.
 */
void error(char *message) 
{
	fprintf(stderr, "%s: %s\n", app_name, message);
	exit(1);
}

/*
 * Obtain checksum via hard disk device.
 */
unsigned char calc_hd_checksum() 
{
	struct hd_driveid id;
	unsigned char *p; 		
	unsigned char c = 0;
	int fd;
	
	fd = open(DEV_HD, O_RDONLY|O_NONBLOCK);
	if (fd == -1) {
		error("error opening hard disk device");
	}
	
	if (!ioctl(fd, HDIO_GET_IDENTITY, &id)) {
		printf("\nHard disk vendor: %s\n", id.model);
		p = id.model;
		while (*p) c += *p++;
		if (!c) c = 0xff;
		printf("Vendor checksum: %x\n", c);
	}
	else {
		close(fd);
		error("no hard disk identification info available");
	}
	close(fd);

	return c;
}

/*
 * Opens I2C device
 */
int open_i2c_device()
{
	int fd;
	
	fd = open(DEV_I2C, O_RDWR);
	if (fd == -1) {
		error("error opening i2c device");
	}
	if (ioctl(fd, I2C_SLAVE, I2C_ADDRESS)) {
		close(fd);
		error("communication with i2c device failed");
	}
	
	return fd;
}

/*
 * Validates HD checksum in EEPROM against calculated checksum.
 */
int validate_hd_checksum()
{
	int sum_check = calc_hd_checksum();
	int fd = open_i2c_device();
	__s32 sum_eeprom = i2c_smbus_read_byte_data(fd, I2C_REGISTER);
	
	if (sum_eeprom == -1) {
		close(fd);
		error("unable to read hard disk checksum via i2c");
	} 
	else {
		printf("EEPROM checksum: %x\n", sum_eeprom);
		if (sum_check == sum_eeprom) {
			printf("Hard disk checksums match!\n");
		} 
		else {
			close(fd);
			if (sum_eeprom == 0) {
				error("hard disk checksum already resetted");
			}
			else {
				error("hard disk checksums differ. Unable to reset");
			}
		}
	}
	close(fd);

	return 1;
}


/*
 * Reset HD checksum via I2C device.
 */
void reset_hd_checksum() 
{
	int fd = open_i2c_device();
	
	if (i2c_smbus_write_byte_data(fd, I2C_REGISTER, 0x00) == -1) {
		close(fd);
		error("unable to write hard disk checksum via i2c");
	}
	close(fd);
}

/* 
 * Print usage text.
 */
void usage() 
{
	printf("\nUsage: hdreset [OPTION]\n");
	printf("\nOptions:");
	printf("\n  -n, --nocheck,  do not perform hard disk checksum validation.");
	printf("\n  -h, --help,     displays this help message.\n\n");
	exit(0);
}

/*
 * Main function.
 */
int main(int argc, char **argv) 
{
	int option_index = 0;
	int c;
	int nocheck = 0;
	int illegal_option = 0;

	static struct option long_options[] = {
		{"nocheck", no_argument, 0, 'n'},
		{"help",    no_argument, 0, 'h'},
		{0, 0, 0, 0}
	};

	app_name = argv[0];
	
	while (!illegal_option && (c = getopt_long (argc, argv, "nh", 
			long_options, &option_index)) != -1) {
		switch (c) {
			case 'n':
				nocheck = 1;
				break;
			case 'h':
				usage();
				break;
			case '?':
				illegal_option = 1;
				break;
			default:
				break;
		}
	}
	
	if (illegal_option) {
		fprintf(stderr, "Try 'hdreset --help' for more information.\n");
	}
	else {
		if (nocheck) {
			reset_hd_checksum();
		} 
		else if (validate_hd_checksum()) {
			reset_hd_checksum();
		}
		printf("\nHard disk checksum resetted successfully.\n\n");
	}
}
